#ifndef __FRAMEWORK__MODULE_H__INCLUDED__
#define __FRAMEWORK__MODULE_H__INCLUDED__

#include <string>
#include <stdexcept>
#include <typeinfo>
#include <iostream>

#ifdef _WINDOWS
#include "_windows.h"
#else
#include <dlfcn.h>
#endif

#include "Exception.h"
#include "Version.h"


FRAMEWORK_BEGIN


DEFINE_EXCEPTION(InvalidModuleException, Exception, "Invalid Module");
DEFINE_EXCEPTION(UndefinedSymbolException, Exception, "Undefined Symbol");


class Module
{
protected:
	static const string	strPrefix;

#ifdef _WINDOWS
	HMODULE			handle;
#else
	void *			handle;
#endif

	const Version	*pVersion;


public:
	enum
	{
#ifdef _WINDOWS
		default_mode
#else
		lazy =			RTLD_LAZY,
		now =			RTLD_NOW,
		global =		RTLD_GLOBAL,
		local =			RTLD_LOCAL,

		default_mode =	lazy
#endif
	};


	template <class ImportedClass> class Allocator
	{
	public:
		typedef Allocator<ImportedClass>	_Type;
		typedef _Type *						_Pointer;


	protected:
		typedef ImportedClass * (*FnNew)();
		typedef ImportedClass * (*FnCopy)(ImportedClass *);
		typedef void (*FnDelete)(ImportedClass *);

		Module		*pModule;

		FnNew		fnNew;
		FnCopy		fnCopy;
		FnDelete	fnDelete;

		uint_t		nAllocated;


	private:
		Allocator<ImportedClass>(_Type &ref) :	pModule(ref.pModule),
												fnNew(ref.fnNew),
												fnCopy(ref.fnCopy),
												fnDelete(ref.fnDelete)
		{	/*	nothing to do	*/	}

		_Type &	operator =(_Type &ref)
		{	/*	nothing to do	*/	}


	public:
		Allocator<ImportedClass>(Module *_pModule) :	pModule(_pModule), nAllocated(0)
		{
			fnNew = (FnNew)pModule->getSymbol(string("new") + ImportedClass::getClassName());
			fnCopy = (FnNew)pModule->getSymbol(string("copy") + ImportedClass::getClassName());
			fnDelete = (FnDelete)pModule->getSymbol(string("delete") + ImportedClass::getClassName());
		}
		inline	~Allocator<ImportedClass>()
		{	if(nAllocated)
				fprintf(stderr, "%s: MEMORY LEAK?: %u instances not deleted\n", __func__, nAllocated);	}

		ImportedClass *	_new()
		{
			ImportedClass	*p = fnNew();
			if(p)
			{	nAllocated++;
				return p;		}
			else
				throw errors::outOfMemory;
		}
		ImportedClass *	_copy(const ImportedClass *pItem)
		{
			ImportedClass	*p = fnCopy(pItem);
			if(p)
			{	nAllocated++;
				return p;		}
			else
				throw errors::outOfMemory;
		}
		void			_delete(ImportedClass *pItem)
		{
			if(nAllocated)
				nAllocated--;
			else
				fprintf(stderr, "%s: deleting %08x, but no instance allocated\n", __func__, pItem);
			fnDelete(pItem);
		}

		inline	_Pointer	__copy()
		{	return this;	}
		inline	void		__delete()
		{	/* nothing to do */	}
	};


	Module(const string &strModule, uint_t nMode = default_mode);
	inline ~Module()
	{
		if(handle)
#ifdef _WINDOWS
			if(!FreeLibrary(handle))
				throw Exception(getLastErrorMessage("failed to free module"));
#else
			if(dlclose(handle) == -1);
				throw Exception(getLastErrorMessage("failed to free module"));
#endif
	}

	inline const Version &	getVersion() const
	{	return *pVersion;	}
	inline void *			getSymbol(const string &strSymbol) const
	{
		void *symbol;

#ifdef _WINDOWS
		if(!(symbol = GetProcAddress(handle, (strPrefix + strSymbol).c_str())))
#else
		if(!(symbol = dlsym(handle, (strPrefix + strSymbol).c_str())))
#endif
			throw UndefinedSymbolException(getLastErrorMessage(strSymbol));

		return symbol;
	}
};


FRAMEWORK_END
#endif
