#ifndef __FRAMEWORK__REFERENCE_H__INCLUDED__
#define __FRAMEWORK__REFERENCE_H__INCLUDED__

#include "Allocator.h"


FRAMEWORK_BEGIN


/*
 * The Reference class template manages other classes, so that a single object
 * can be used from multiple "locations", without the need to keep track of
 * the references to properly delete the instance. The class counts the
 * references to an objects and destroys it when no more remain.
 *
 * For allocating and deallocating resources, Allocator classes are used. This
 * could be template allocating memory using new, or a class creating (and
 * destroying) a mutex,...
 */
template <	class Item,
			class Id = void,
			class Allocator = Allocator<Item, Id>,
			bool STATIC_ALLOCATOR = false>
	class Reference
{
public:
	typedef Reference<	Item,
						Id,
						Allocator,
						STATIC_ALLOCATOR	>
			_Type;

protected:
	Allocator	*pAllocator;
	Item		*pItem;
	uint_t		*pnReferences;

	void		__init_dynamic(const Allocator *_pAllocator, const bool bCreate)
	{
		if(_pAllocator)
			pAllocator = _pAllocator->__copy();
		else if(bCreate)
			pAllocator = Allocator::__new();
		else
			pAllocator = NULL;
	}
	inline void	_init(const Allocator *_pAllocator, const bool bCreate = false)
	{
		if(STATIC_ALLOCATOR)
			pAllocator = const_cast<Allocator *>(_pAllocator);
		else
			__init_dynamic(_pAllocator, bCreate);
	}

	void		_init(Item *_pItem, const Allocator *_pAllocator)
	{
		_init(_pAllocator, _pItem != NULL);

		pItem = _pItem;
		if(pItem)
		{
			pnReferences = new uint_t;
			if(!pnReferences)
				throw OutOfMemoryError();
			*pnReferences = 1;
		}
		else
			pnReferences = NULL;
	}
	void		_init(const Item *_pItem, const Allocator *_pAllocator)
	{
		_init(_pAllocator, _pItem != NULL);

		if(_pItem)
		{
			pItem = pAllocator->_copy(_pItem);
			pnReferences = new uint_t;
			if(!pnReferences)
				throw OutOfMemoryError();
			*pnReferences = 1;
		}
		else
			pnReferences = NULL;
	}

	inline void	__release_dynamic()
	{	if(pAllocator)
			pAllocator->__delete();	}
	void		_release(bool bAll = false)
	{
		if(pnReferences)
		{
			(*pnReferences)--;
			if(!(*pnReferences))
			{
				if(pItem)
					pAllocator->_delete(pItem);
				if(bAll)
					delete pnReferences;
			}
		}
		if(!STATIC_ALLOCATOR && bAll)
			__release_dynamic();
	}

	inline void	__needAllocator_dynamic()
	{	if(!pAllocator)
			pAllocator = Allocator::__new();	}
	inline void	_needAllocator()
	{
		if(STATIC_ALLOCATOR)
		{
			if(!pAllocator)
				throw NullPointerException("*STATIC* allocator is NULL");
		}
		else
			__needAllocator_dynamic();
	}

	inline _Type	_copy_dynamic()
	{	_needAllocator();
		return pItem ? _Type(pAllocator->_copy(pItem), pAllocator->__copy()) : _Type();	}
	inline _Type	_copy_dynamic() const
	{
		if(!pAllocator)
			throw NullPointerException("constant allocator is NULL");
		return pItem ? _Type(pAllocator->_copy(pItem), pAllocator->__copy()) : _Type();
	}
	inline _Type	_copy_static() const
	{
		if(!pAllocator)
			throw NullPointerException("*STATIC* allocator is NULL");
		return pItem ? _Type(pAllocator->_copy(pItem), pAllocator) : _Type();
	}

public:
	inline	Reference<	Item,
						Id,
						Allocator,
						STATIC_ALLOCATOR	>(const Allocator *_pAllocator = NULL) :
		pItem(NULL),
		pnReferences(NULL)
	{	_init(_pAllocator);	}
	inline	Reference<	Item,
						Id,
						Allocator,
						STATIC_ALLOCATOR	>(Item *_pItem, const Allocator *_pAllocator = NULL)
	{	_init(_pItem, _pAllocator);	}
	inline	Reference<	Item,
						Id,
						Allocator,
						STATIC_ALLOCATOR	>(const Item *_pItem, const Allocator *_pAllocator = NULL)
	{	_init(_pItem, _pAllocator);	}
	inline	Reference<	Item,
						Id,
						Allocator,
						STATIC_ALLOCATOR	>(const Item &item, const Allocator *_pAllocator = NULL)
	{	_init(&item, _pAllocator);	}
	Reference<	Item,
				Id,
				Allocator,
				STATIC_ALLOCATOR	>(const _Type &ref) :
		pAllocator((STATIC_ALLOCATOR || !ref.pAllocator) ? ref.pAllocator : ref.pAllocator->__copy()),
		pItem(ref.pItem),
		pnReferences(ref.pnReferences)
	{	if(pnReferences)
			(*pnReferences)++;	}
	virtual	~Reference<	Item,
						Id,
						Allocator,
						STATIC_ALLOCATOR	>()
	{	_release(true);	}
	
	_Type &			operator =(Item *_pItem)
	{
		if(pItem != _pItem)
		{
			if(_pItem)
			{
				if(pItem)
				{
					(*pnReferences)--;
					if(!(*pnReferences))
						pAllocator->_delete(pItem);
					else if(*pnReferences)
						pnReferences = new uint_t;
				}
				else
				{
					_needAllocator();
					if(!pnReferences)
						pnReferences = new uint_t;
				}

				pItem = _pItem;
				(*pnReferences) = 1;
			}
			else
			{
				_release();
				pItem = NULL;
			}
		}

		return *this;
	}
	inline _Type &	operator =(const Item *_pItem)
	{
		if(pItem != _pItem)
		{
			_needAllocator();
			*this = pAllocator->_copy(pItem);
		}
		return *this;
	}
	inline _Type &	operator =(const Item &item)
	{	return operator =(&item);	}
	_Type &	operator =(const _Type &ref)
	{
		if(&ref != this)
		{
			_release(true);

			_init(ref.pAllocator);
			pItem = ref.pItem;
			pnReferences = ref.pnReferences;
			if(pItem)
				(*pnReferences)++;
		}

		return *this;
	}

	inline Item *		getPointer()
	{	return pItem;	}
	inline const Item *	getPointer() const
	{	return pItem;	}

	inline uint_t	getReferenceCount() const
	{	return pnReferences ? *pnReferences : 0;	}

	inline bool		isEmpty() const
	{	return pItem == NULL;	}

	inline _Type &	create()
	{	_needAllocator();
		return *this = pAllocator->_new();	}
	inline _Type &	create(const Id *pId)
	{	_needAllocator();
		return *this = pAllocator->_new(pId);	}
	inline _Type	copy() const
	{
		if(STATIC_ALLOCATOR)
			return _copy_static();
		else
			return _copy_dynamic();
	}

	inline Item *		operator ->()
	{	return pItem;	}
	inline const Item *	operator ->() const
	{	return pItem;	}

	inline bool	operator <(const _Type &ref) const
	{	return pItem < ref.pItem;	}
	inline bool	operator <=(const _Type &ref) const
	{	return pItem <= ref.pItem;	}
	inline bool	operator >(const _Type &ref) const
	{	return pItem > ref.pItem;	}
	inline bool	operator >=(const _Type &ref) const
	{	return pItem >= ref.pItem;	}
	inline bool	operator ==(const Item *pItem) const
	{	return pItem == pItem;	}
	inline bool	operator ==(const Item &item) const
	{	return pItem == &item;	}
	inline bool	operator ==(const _Type &ref) const
	{	return pItem == ref.pItem;	}
	inline bool	operator !=(const Item *pItem) const
	{	return pItem != pItem;	}
	inline bool	operator !=(const Item &item) const
	{	return pItem != &item;	}
	inline bool	operator !=(const _Type &ref) const
	{	return pItem != ref.pItem;	}

	inline operator *()
	{	return *pItem;	}
	inline operator *() const
	{	return *pItem;	}
};


FRAMEWORK_END


#include "Exception.h"


#endif
