#ifndef _MAP_H
#define _MAP_H

#include <vector>

#include "Texture.h"
#include "Object3.h"
#include <graphics/texload.h>


GRAPHICS_BEGIN

#define TERRAIN_CATHEGORY	0
#define TERRAIN_TYPE		0


// Das ist unsere Map-Klasse
class Map :	public GlItem
{
protected:
	enum
	{
		MIN_BLOCKSIZE_EXPONENT	= 5,
		MAX_BLOCKSIZE_EXPONENT	= 7,
		MAX_LOD					= 4,

		LOD_COUNT				= MAX_LOD + 1
	};


	class Data
	{
	protected:
		static inline uint8_t	_getBlockSizeExponent(const uint8_t _nExponent)
		{
			return	_nExponent < MIN_BLOCKSIZE_EXPONENT ?
						MIN_BLOCKSIZE_EXPONENT
					:
						( _nExponent > MAX_BLOCKSIZE_EXPONENT ?
							MAX_BLOCKSIZE_EXPONENT
						:
							_nExponent							);
		}


	public:
		// global tarrain config
		static bool			bFadeTextures;
		// light vector, ambient light
		static GLfloat		fLightX,
							fLightY,
							fLightZ,
							fLightMin;

		// terrain textures
		Texture::Reference	texture0,
							texture1;

		// the height map
		uint16_t			*pHeightMap;

		// terrain parameters
		const uint_t		nMapSizeX,
							nMapSizeZ,
							nHeightMapX;
		const uint8_t		nBlockSizeExponent;
		const GLfloat		fScaleX,
							fScaleY,
							fScaleZ;

		// global terrain data (triange strip indices,...)
		uint_t				nStartIndex[LOD_COUNT + 1],
							nLineLength[LOD_COUNT];
		GLushort			*pIndices,
							*pCornerBottomLeft,
							*pCornerBottomRight,
							*pCornerTopRight,
							*pCornerTopLeft,
							nCornerIndex[4],
							nInnerCornerIndex[LOD_COUNT << 2];
		GLuint				nIndexId;


		Data(	const uint_t &_nMapSizeX,
				const uint_t &_nMapSizeZ,
				const uint8_t &_nBlockSizeExponent,
				const GLfloat &_fScaleX = 1.0f,
				const GLfloat &_fScaleY = 1.0f / 512.0f,
				const GLfloat &_fScaleZ = 1.0f			);
		~Data();


		void			generateStrips();


		inline GLushort	getBlockSize() const
		{	return (GLushort)1 << nBlockSizeExponent;	}

		inline GLushort	getLoDVertexIndex(const uint8_t &nLoD)
		{	// sum(2^MAX_LOD - 2^(MAX_LOD - i), i, 0, nLevel - 1)
			return ((((GLushort)nLoD) << MAX_LOD) - ((GLushort)1 << nLoD) + 1) << (nBlockSizeExponent - MAX_LOD);	}
		inline GLushort	getLoDVertexCount()
		{	return getLoDVertexIndex(LOD_COUNT);	}

		inline GLushort	getLoDCornerIndex(const uint8_t &nLoD)
		{	// sum(MAX_LOD - i + 1, i, 0, nLevel - 1)
			return LOD_COUNT * (GLushort)nLoD - (((GLushort)nLoD * nLoD - nLoD) >> 1);	}
		inline GLushort	getLoDCornerCount()
		{	return getLoDCornerIndex(LOD_COUNT);	}
	};


public:
	class Block :	public Object3
	{
	protected:
		// map data
		Data			*pMapData;
		// this block's position
		const uint_t	nBlockX,
						nBlockZ;

		// neighbours
		const Block		*pLeft,
						*pTop,
						*pRight,
						*pBottom;

		uint8_t			nLoD;
		float			fRoughness[LOD_COUNT];

		GLfloat			*pVertices,
						*pLightMap;
		uint_t			nVertices;
		GLuint			nVboId,
						nLightMapId;


		void			_generateVertexNormals();
		void			_generateAlpha();


		virtual void	_load();
		virtual void	_remove();

		virtual void	_glLoad();
		virtual void	_glRemove();

		virtual void	_glRender();


	public:
		Block(Map *parent, const uint_t &blockx, const uint_t &blockz);
		virtual ~Block();


		inline void		neighbours(const Block *left, const Block *top, const Block *right, const Block *bottom)
		{	pLeft = left;	pTop = top;	pRight = right; pBottom = bottom;	}
		inline void		bottom(const Block *_bottom)
		{	pBottom = _bottom;	}
		inline void		right(const Block *_right)
		{	pRight = _right;	}
		inline void		top(const Block *_top)
		{	pTop = _top;	}
		inline void		left(const Block *_left)
		{	pLeft = _left;	}

		virtual void	cameraPosition(const Vector &camera);


		virtual void	glReload();

		virtual void	glBind();
	};


protected:
	// all the map data
	Data				data;

	// height map (file name, copy in memory)
    string				strFile;

	// terrain blocks
	vector<_Object3 *>	blocks;


	virtual void	_load();
	virtual void	_remove();

	virtual void	_glLoad();
	virtual void	_glRemove();


public:
	// Unser Map-Konstruktor, mit bergabeparamtern.
	Map(const string &_strFile,
		const uint_t &mapsixex,
		const uint_t &mapsizez,
		const uint8_t nBlockSizeExponent,
		const float &scalex = 1.0f,
		const float &scaley = 1.0f / 512.0f,
		const float &scalez = 1.0f			);
	~Map();


	inline void toggleTextureFading()
	{	data.bFadeTextures = !data.bFadeTextures;		}
	inline void setLightVector(const float &x, const float &y, const float &z, const float &min = 0.0f)
	{	data.fLightX = -x;	data.fLightY = -y;	data.fLightZ = -z;	data.fLightMin = min;	}

	inline void	textures(Texture::Reference &_newtexture0, Texture::Reference &_newtexture1)
	{	data.texture0 = _newtexture0;
		data.texture1 = _newtexture1;	}
	

	// grenzen dieser map
	inline float	getMinX() const
	{	return data.nMapSizeX / -2.0f * data.fScaleX;	}
	inline float	getMaxX() const
	{	return data.nMapSizeX / 2.0f * data.fScaleX;	}
	inline float	getMinZ() const
	{	return data.nMapSizeZ / -2.0f * data.fScaleZ;	}
	inline float	getMaxZ() const
	{	return data.nMapSizeZ / 2.0f * data.fScaleZ;	}


	// interpolation der hhenwerte
	float			_heightAt(float x, float z) const;
	inline float	heightAt(const float _x, const float _z) const
	{	const float	x = _x / data.fScaleX + (data.nMapSizeX + 1.0f) / 2.0f,
					z = _z / data.fScaleZ + (data.nMapSizeZ + 1.0f) / 2.0f;
		return	_heightAt(x > data.nMapSizeX ? data.nMapSizeX : x, z > data.nMapSizeZ ? data.nMapSizeZ : z) *
				data.fScaleY;																					}
	// berechnung der steigung
	float			_slopeAt(float x, float z, const float &dx, const float &dz) const;
	inline float	slopeAt(const float _x, const float _z, const float &dx = 1.0f, const float &dz = 1.0f) const
	{	return _slopeAt(	_x / data.fScaleX + (data.nMapSizeX + 1.0f) / 2.0f,
							_z / data.fScaleZ + (data.nMapSizeZ + 1.0f) / 2.0f, dx, dz) *
				data.fScaleY;																}


	inline operator vector<_Object3 *> &()
	{	return blocks;	}
	inline operator const vector<_Object3 *> &() const
	{	return blocks;	}
};


GRAPHICS_END
#endif
