#ifndef __GRAPHICS__OBJECT_3_H__INCLUDED__
#define __GRAPHICS__OBJECT_3_H__INCLUDED__

#include <map>

#include <framework/Reference.h>

#include "Object.h"
#include "Renderable.h"
#include "Material.h"
#include <graphics/texload.h>


GRAPHICS_BEGIN

class _Object3 :	public Object
{
public:
	typedef _Vector3										Vector;
	typedef _Vector3D										VectorD;

	typedef	multimap<float, Renderable *, greater<float> >	RenderMap;


	enum Boundary
	{
		none,
		sphere,
		box
	};


protected:
	// bounding object data
	Boundary	boundaryType;
	GLfloat		*pBoundary;
	bool		bCanCollide;

	bool		bVisibleRange;
	GLfloat		fVisibleRange,
				fCameraDistance;

	// object type and data
	uint_t		nCathegoryID,
				nTypeID;
	void		*pTypeData;


	// collision detection
	static bool		_collides_spheres(_Object3 &o1, _Object3 &o2);
	static bool		_collides_sphere_box(_Object3 &o1, _Object3 &o2);
	static bool		_collides_boxes(_Object3 &o1, _Object3 &o2);

	// collision detection, more detailed than just with the bounding object
	virtual bool	_collides(_Object3 &o);


public:
	_Object3(const bool _bCanCollide = false);
	virtual ~_Object3();


	// set the bounding object
	void			boundingBox(	const GLfloat &sx, const GLfloat &sy, const GLfloat &sz,
									const GLfloat &x = 0.0f, const GLfloat &y = 0.0f, const GLfloat &z = 0.0f	);
	void			boundingSphere(	const GLfloat &radius,
									const GLfloat &x = 0.0f, const GLfloat &y = 0.0f, const GLfloat &z = 0.0f	);

	// return the bounding object
	inline const Boundary &	getBoundaryType() const
	{	return boundaryType;	}
	inline const GLfloat *	getBoundary() const
	{	return pBoundary;		}

	void					getBoundingBox(Vector *box) const;
	GLfloat					getBoundingRadiusSquared() const;
	GLfloat					getBoundingRadius() const;

	inline void				visibleRange(const GLfloat &range)
	{	bVisibleRange = true;
		fVisibleRange = range * range;	}
	inline void				noVisibleRange()
	{	bVisibleRange = false;	}
	inline bool				useVisibleRange() const
	{	return bVisibleRange;	}
	inline GLfloat			getVisibleRange() const
	{	return sqrt(fVisibleRange);	}
	inline GLfloat			getVisibleRangeSquared() const
	{	return fVisibleRange;	}


	// save the camera position
	virtual void	cameraPosition(const Vector &camera);

	// collision detection
	inline void		canCollide(const bool _bCanCollide)
	{	bCanCollide = _bCanCollide;	}
	inline bool		canCollide() const
	{	return bCanCollide;	}

	bool			collides(_Object3 &o);

	
	// object type handling
	inline void		type(const uint_t &_nCathegoryID, const uint_t &_nTypeId, void *_pTypeData = NULL)
	{	nCathegoryID = _nCathegoryID;
		nTypeID = _nTypeId;
		pTypeData = _pTypeData;			}

	inline uint_t		getCathegoryID() const
	{	return nCathegoryID;	}
	inline uint_t		getTypeID() const
	{	return nTypeID;	}
	inline void *		getTypeData()
	{	return pTypeData;	}
	inline const void *	getTypeData() const
	{	return pTypeData;	}


	// transformations
	inline void	preRotate(const GLfloat &x, const GLfloat &y, const GLfloat &z, const GLfloat &theta)
	{	tmLocal.preRotate(x, y, z, theta);	}
	inline void	preRotate(Vector axis, const GLfloat &theta)
	{	axis.normalise();
		preRotate(axis.getX(), axis.getY(), axis.getZ(), theta);	}
	inline void	postRotate(const GLfloat &x, const GLfloat &y, const GLfloat &z, const GLfloat &theta)
	{	tmLocal.postRotate(x, y, z, theta);	}
	inline void	postRotate(Vector axis, const GLfloat &theta)
	{	axis.normalise();
		postRotate(axis.getX(), axis.getY(), axis.getZ(), theta);	}

	inline void	preRotateX(GLfloat phi)
	{	tmLocal.preRotateX(phi);	}
	inline void	postRotateX(GLfloat phi)
	{	tmLocal.postRotateX(phi);	}
	inline void	preRotateY(GLfloat theta)
	{	tmLocal.preRotateY(theta);	}
	inline void	postRotateY(GLfloat theta)
	{	tmLocal.postRotateY(theta);	}
	inline void	preRotateZ(GLfloat psi)
	{	tmLocal.preRotateZ(psi);	}
	inline void	postRotateZ(GLfloat psi)
	{	tmLocal.postRotateZ(psi);	}

	inline void	preTranslate(const GLfloat &x, const GLfloat &y, const GLfloat &z)
	{	tmLocal.preTranslate(x, y, z);	}
	inline void	preTranslate(const Vector &vDifference)
	{	preTranslate(vDifference.getX(), vDifference.getY(), vDifference.getZ());	}
	inline void	postTranslate(const GLfloat &x, const GLfloat &y, const GLfloat &z)
	{	tmLocal.postTranslate(x, y, z);	}
	inline void	postTranslate(const Vector &vDifference)
	{	postTranslate(vDifference.getX(), vDifference.getY(), vDifference.getZ());	}

	inline void	preScale(const GLfloat &sx, const GLfloat &sy, const GLfloat &sz)
	{	tmLocal.preScale(sx, sy, sz);	}
	inline void	postScale(const GLfloat &sx, const GLfloat &sy, const GLfloat &sz)
	{	tmLocal.postScale(sx, sy, sz);	}


	// get functions
	inline void		getPosition(Vector &v) const
	{	v.set(	tmFinal.getElement(0, 3),
				tmFinal.getElement(1, 3),
				tmFinal.getElement(2, 3)	);	}
	inline void		getPosition(VectorD &v) const
	{	v.set(	tmFinal.getElement(0, 3),
				tmFinal.getElement(1, 3),
				tmFinal.getElement(2, 3)	);	}
	inline Vector	getPosition() const
	{	return Vector(	tmFinal.getElement(0, 3),
						tmFinal.getElement(1, 3),
						tmFinal.getElement(2, 3)	);	}
	inline VectorD	getPositionD()
	{	return VectorD(	tmFinal.getElement(0, 3),
						tmFinal.getElement(1, 3),
						tmFinal.getElement(2, 3)	);	}
	inline GLfloat	getPositionX() const
	{	return tmFinal.getElement(0, 3);	}
	inline GLfloat	getPositionY() const
	{	return tmFinal.getElement(1, 3);	}
	inline GLfloat	getPositionZ() const
	{	return tmFinal.getElement(2, 3);	}

	inline void		getForward(Vector &v) const
	{	v.set(	-tmFinal.getElement(0, 2),
				-tmFinal.getElement(1, 2),
				-tmFinal.getElement(2, 2)	);	}
	inline void		getForward(VectorD &v) const
	{	v.set(	-tmFinal.getElement(0, 2),
				-tmFinal.getElement(1, 2),
				-tmFinal.getElement(2, 2)	);	}
	inline Vector	getForward() const
	{	return Vector(	-tmFinal.getElement(0, 2),
						-tmFinal.getElement(1, 2),
						-tmFinal.getElement(2, 2)	);	}
	inline VectorD	getForwardD() const
	{	return VectorD(	-tmFinal.getElement(0, 2),
						-tmFinal.getElement(1, 2),
						-tmFinal.getElement(2, 2)	);	}
	inline void		getUp(Vector &v) const
	{	v.set(	tmFinal.getElement(0, 1),
				tmFinal.getElement(1, 1),
				tmFinal.getElement(2, 1)	);	}
	inline void		getUp(VectorD &v) const
	{	v.set(	tmFinal.getElement(0, 1),
				tmFinal.getElement(1, 1),
				tmFinal.getElement(2, 1)	);	}
	inline Vector	getUp() const
	{	return Vector(	tmFinal.getElement(0, 1),
						tmFinal.getElement(1, 1),
						tmFinal.getElement(2, 1)	);	}
	inline VectorD	getUpD() const
	{	return VectorD(	tmFinal.getElement(0, 1),
						tmFinal.getElement(1, 1),
						tmFinal.getElement(2, 1)	);	}
	inline void		getStrafe(Vector &v) const
	{	v.set(	tmFinal.getElement(0, 0),
				tmFinal.getElement(1, 0),
				tmFinal.getElement(2, 0)	);	}
	inline void		getStrafe(VectorD &v) const
	{	v.set(	tmFinal.getElement(0, 0),
				tmFinal.getElement(1, 0),
				tmFinal.getElement(2, 0)	);	}
	inline Vector	getStrafe()
	{	return Vector(	tmFinal.getElement(0, 0),
						tmFinal.getElement(1, 0),
						tmFinal.getElement(2, 0)	);	}
	inline VectorD	getStrafeD()
	{	return VectorD(	tmFinal.getElement(0, 0),
						tmFinal.getElement(1, 0),
						tmFinal.getElement(2, 0)	);	}


	inline void		getLocalPosition(Vector &v) const
	{	v.set(	tmLocal.getElement(0, 3),
				tmLocal.getElement(1, 3),
				tmLocal.getElement(2, 3)	);	}
	inline void		getLocalPosition(VectorD &v) const
	{	v.set(	tmLocal.getElement(0, 3),
				tmLocal.getElement(1, 3),
				tmLocal.getElement(2, 3)	);	}
	inline Vector	getLocalPosition() const
	{	return Vector(	tmLocal.getElement(0, 3),
						tmLocal.getElement(1, 3),
						tmLocal.getElement(2, 3)	);	}
	inline VectorD	getLocalPositionD()
	{	return VectorD(	tmLocal.getElement(0, 3),
						tmLocal.getElement(1, 3),
						tmLocal.getElement(2, 3)	);	}
	inline GLfloat	getLocalPositionX() const
	{	return tmLocal.getElement(0, 3);	}
	inline GLfloat	getLocalPositionY() const
	{	return tmLocal.getElement(1, 3);	}
	inline GLfloat	getLocalPositionZ() const
	{	return tmLocal.getElement(2, 3);	}

	inline void		getLocalForward(Vector &v) const
	{	v.set(	-tmLocal.getElement(0, 2),
				-tmLocal.getElement(1, 2),
				-tmLocal.getElement(2, 2)	);	}
	inline void		getLocalForward(VectorD &v) const
	{	v.set(	-tmLocal.getElement(0, 2),
				-tmLocal.getElement(1, 2),
				-tmLocal.getElement(2, 2)	);	}
	inline Vector	getLocalForward() const
	{	return Vector(	-tmLocal.getElement(0, 2),
						-tmLocal.getElement(1, 2),
						-tmLocal.getElement(2, 2)	);	}
	inline VectorD	getLocalForwardD() const
	{	return VectorD(	-tmLocal.getElement(0, 2),
						-tmLocal.getElement(1, 2),
						-tmLocal.getElement(2, 2)	);	}
	inline void		getLocalUp(Vector &v) const
	{	v.set(	tmLocal.getElement(0, 1),
				tmLocal.getElement(1, 1),
				tmLocal.getElement(2, 1)	);	}
	inline void		getLocalUp(VectorD &v) const
	{	v.set(	tmLocal.getElement(0, 1),
				tmLocal.getElement(1, 1),
				tmLocal.getElement(2, 1)	);	}
	inline Vector	getLocalUp() const
	{	return Vector(	tmLocal.getElement(0, 1),
						tmLocal.getElement(1, 1),
						tmLocal.getElement(2, 1)	);	}
	inline VectorD	getLocalUpD() const
	{	return VectorD(	tmLocal.getElement(0, 1),
						tmLocal.getElement(1, 1),
						tmLocal.getElement(2, 1)	);	}
	inline void		getLocalStrafe(Vector &v) const
	{	v.set(	tmLocal.getElement(0, 0),
				tmLocal.getElement(1, 0),
				tmLocal.getElement(2, 0)	);	}
	inline void		getLocalStrafe(VectorD &v) const
	{	v.set(	tmLocal.getElement(0, 0),
				tmLocal.getElement(1, 0),
				tmLocal.getElement(2, 0)	);	}
	inline Vector	getLocalStrafe()
	{	return Vector(	tmLocal.getElement(0, 0),
						tmLocal.getElement(1, 0),
						tmLocal.getElement(2, 0)	);	}
	inline VectorD	getLocalStrafeD()
	{	return VectorD(	tmLocal.getElement(0, 0),
						tmLocal.getElement(1, 0),
						tmLocal.getElement(2, 0)	);	}


	inline void		getGlPosition(Vector &v) const
	{	v.set(	tmGL.getElement(0, 3),
				tmGL.getElement(1, 3),
				tmGL.getElement(2, 3)	);	}
	inline void		getGlPosition(VectorD &v) const
	{	v.set(	tmGL.getElement(0, 3),
				tmGL.getElement(1, 3),
				tmGL.getElement(2, 3)	);	}
	inline Vector	getGlPosition() const
	{	return Vector(	tmGL.getElement(0, 3),
						tmGL.getElement(1, 3),
						tmGL.getElement(2, 3)	);	}
	inline VectorD	getGlPositionD()
	{	return VectorD(	tmGL.getElement(0, 3),
						tmGL.getElement(1, 3),
						tmGL.getElement(2, 3)	);	}
	inline GLfloat	getGlPositionX() const
	{	return tmGL.getElement(0, 3);	}
	inline GLfloat	getGlPositionY() const
	{	return tmGL.getElement(1, 3);	}
	inline GLfloat	getGlPositionZ() const
	{	return tmGL.getElement(2, 3);	}

	inline void		getGlForward(Vector &v) const
	{	v.set(	-tmGL.getElement(0, 2),
				-tmGL.getElement(1, 2),
				-tmGL.getElement(2, 2)	);	}
	inline void		getGlForward(VectorD &v) const
	{	v.set(	-tmGL.getElement(0, 2),
				-tmGL.getElement(1, 2),
				-tmGL.getElement(2, 2)	);	}
	inline Vector	getGlForward() const
	{	return Vector(	-tmGL.getElement(0, 2),
						-tmGL.getElement(1, 2),
						-tmGL.getElement(2, 2)	);	}
	inline VectorD	getGlForwardD() const
	{	return VectorD(	-tmGL.getElement(0, 2),
						-tmGL.getElement(1, 2),
						-tmGL.getElement(2, 2)	);	}
	inline void		getGlUp(Vector &v) const
	{	v.set(	tmGL.getElement(0, 1),
				tmGL.getElement(1, 1),
				tmGL.getElement(2, 1)	);	}
	inline void		getGlUp(VectorD &v) const
	{	v.set(	tmGL.getElement(0, 1),
				tmGL.getElement(1, 1),
				tmGL.getElement(2, 1)	);	}
	inline Vector	getGlUp() const
	{	return Vector(	tmGL.getElement(0, 1),
						tmGL.getElement(1, 1),
						tmGL.getElement(2, 1)	);	}
	inline VectorD	getGlUpD() const
	{	return VectorD(	tmGL.getElement(0, 1),
						tmGL.getElement(1, 1),
						tmGL.getElement(2, 1)	);	}
	inline void		getGlStrafe(Vector &v) const
	{	v.set(	tmGL.getElement(0, 0),
				tmGL.getElement(1, 0),
				tmGL.getElement(2, 0)	);	}
	inline void		getGlStrafe(VectorD &v) const
	{	v.set(	tmGL.getElement(0, 0),
				tmGL.getElement(1, 0),
				tmGL.getElement(2, 0)	);	}
	inline Vector	getGlStrafe()
	{	return Vector(	tmGL.getElement(0, 0),
						tmGL.getElement(1, 0),
						tmGL.getElement(2, 0)	);	}
	inline VectorD	getGlStrafeD()
	{	return VectorD(	tmGL.getElement(0, 0),
						tmGL.getElement(1, 0),
						tmGL.getElement(2, 0)	);	}

	virtual void	glGetAll(RenderMap &rm, const Vector &vCameraPosition) = 0;
	virtual void	glGetVisible(RenderMap &rm, Camera *pCamera, const Vector &vCameraPosition) = 0;
};


class Object3 :	public _Object3,
				public Renderable
{
protected:
	Material		*pMaterial;

	static GLfloat		fCausticScale,
						fCausticTime;

	static int			nCausticIncrement;

	static GLuint		uiaCausticTextures[32];

	bool				bRenderCaustics;

	// render the object
	virtual void	_glRender() = 0;


public:
	inline Object3(const bool _bCanCollide = 0, Material *_pMaterial = NULL, bool _bRenderCaustics = 1) :
		_Object3(_bCanCollide),
		bRenderCaustics(_bRenderCaustics),
		pMaterial(_pMaterial),
		Renderable()
	{	/* nothing to do */ }


	inline void		material(Material *_pMaterial)
	{	pMaterial = _pMaterial;	}

	static inline GLfloat advanceCausticTime(float fTime)
	{
		fCausticTime += fTime/2.0f;

		while(fCausticTime >= 1)
			fCausticTime -= 1;

		return fCausticTime;
	}

	// enable/disable caustics
	inline void	setCaustics(bool _bRenderCaustics)
	{	bRenderCaustics = _bRenderCaustics;		}

	virtual void	glGetAll(RenderMap &rm, const Vector &vCameraPosition);
	virtual void	glGetVisible(RenderMap &rm, Camera *pCamera, const Vector &vCameraPosition);

	// load caustic textures
	static void	loadCausticTextures();

	// load, bind and render the object
	virtual void	glRender();
};


GRAPHICS_END
#endif
