#ifndef __GRAPHICS__TEXTURE_H__INCLUDED__
#define __GRAPHICS__TEXTURE_H__INCLUDED__

#include <framework/ClassManager.h>
#include <framework/ObjectManager.h>

#include "GlItem.h"
#include "Colour.h"


GRAPHICS_BEGIN


class Texture	: public GlItem
{
public:
	class Id
	{
	public:
		class ClassId
		{
		protected:
			const string	extension;

		public:
			inline ClassId(const string &_extension) :	extension(_extension)
			{	/* nothing to do */	}

			inline const string &	getExtension() const
			{	return extension;	}

			inline bool	operator <(const ClassId &classId) const
			{	return extension < classId.extension;	}
			inline bool	operator <=(const ClassId &classId) const
			{	return extension <= classId.extension;	}
			inline bool	operator >(const ClassId &classId) const
			{	return extension > classId.extension;	}
			inline bool	operator >=(const ClassId &classId) const
			{	return extension >= classId.extension;	}
			inline bool	operator ==(const ClassId &classId) const
			{	return extension == classId.extension;	}
			inline bool	operator !=(const ClassId &classId) const
			{	return extension != classId.extension;	}
		};

	protected:
		const ClassId	classId;

		static inline string	_getExtension(const string &file)
		{	size_t	index = file.find_last_of('.');
			return index == string::npos ? string() : file.substr(index + 1);	}

	public:
		const string	file;
		const bool		crop,
						clamp,
						mipmap;

		Id(	const string &_file,
			const bool _crop = true,
			const bool _clamp = false,
			const bool _mipmap = false	);

		inline const string &	extension() const
		{	return classId.getExtension();	}

		inline bool	operator <(const Id &id) const
		{	return file < id.file;	}
		inline bool	operator <=(const Id &id) const
		{	return file <= id.file;	}
		inline bool	operator >(const Id &id) const
		{	return file > id.file;	}
		inline bool	operator >=(const Id &id) const
		{	return file >= id.file;	}
		inline bool	operator ==(const Id &id) const
		{	return file == id.file;	}
		inline bool	operator !=(const Id &id) const
		{	return file != id.file;	}

		operator const ClassId &() const
		{	return classId;	}
	};

	class Manager :	public ObjectManager<	Texture,
											Id,
											ClassManager<Texture, Id, Id::ClassId>,
											true,
											IAllocatorBase<Texture, Id>,
											false									>
	{
	public:
		typedef ObjectManager<	Texture,
								Id,
								ClassManager<Texture, Id, Id::ClassId>,
								true,
								IAllocatorBase<Texture, Id>,
								false									>
				_Base;

		typedef ClassManager<Texture, Id, Id::ClassId>	ClassManager;
		typedef IAllocator<Texture, Id>					Allocator;

	protected:
		Allocator	*pTGAAllocator;

		void	_init();

	public:
		inline Manager()
		{	_init();	}
		inline Manager(Manager &m) :	_Base(m)
		{	_init();	}
		virtual ~Manager();

		inline ClassManager &	classes()
		{	return allocator;	}
	};

	typedef	Manager::Reference	Reference;
	typedef	FloatColour			Colour;

	typedef Colour (Texture::*fnGetTexel)(const uint_t &index) const;


protected:
	typedef	Manager::Allocator	Allocator;


	static Manager	manager;
	static bool		bGlobalMipmap;


	GLuint		nID;			// opengl texture id
	GLenum		nType;			// texel type (GL_RGB or GL_RGBA)
	GLuint		nWidth,			// the texture's width (must be a power of 2)
				nHeight,		// the texture's height (must be a power of 2)
				nTexelSize;
	GLfloat		*pImageData;	// pointer to the image data (the texels)

	bool		bClamp,
				bMipmap;


	Texture(	const bool bClamp = false,
				const bool bMipmap = true	);

	void			_init_type();

	virtual	void	_remove();


	/****************************************************************/
	// OpenGL methods
	/****************************************************************/
	virtual void	_glLoad();					// pass the texture to OpenGL
	virtual void	_glRemove();				// removes the texture from OpenGL

	void			_glApplyParameters() const;	// set texture options


	fnGetTexel	_getTexel;
	Colour		_getTexel_GL_RED(const uint_t index) const;
	Colour		_getTexel_GL_GREEN(const uint_t index) const;
	Colour		_getTexel_GL_BLUE(const uint_t index) const;
	Colour		_getTexel_GL_ALPHA(const uint_t index) const;
	Colour		_getTexel_GL_LUMINANCE(const uint_t index) const;
	Colour		_getTexel_GL_LUMINANCE_ALPHA(const uint_t index) const;
	Colour		_getTexel_GL_RGB(const uint_t index) const;
	Colour		_getTexel_GL_BGR_EXT(const uint_t index) const;
	Colour		_getTexel_GL_RGBA(const uint_t index) const;
	Colour		_getTexel_GL_BGRA_EXT(const uint_t index) const;
	Colour		_getTexel_unknown(const uint_t index) const;


public:
	virtual	~Texture();

	static inline void		add(const Id::ClassId &classId, Allocator *pAllocator)
	{	manager.classes().add(classId, pAllocator);	}

	// creates a texture object, return NULL on failure
	static inline Reference	create(	const string &file,
									const bool crop = true,
									const bool clamp = false,
									const bool mipmap = false	)
	{	return manager.createReference(Id(file, crop, clamp, mipmap));	}

	static inline void		clear()
	{	manager.clear();	}
	static inline void		clean()
	{	manager.clean();	}


	static inline void		enableGlobalMipmapping(const bool enable)
	{	bGlobalMipmap = enable;	}

	static uint_t			powerOf2(const uint_t upperBound);

	/****************************************************************/
	// get methods
	/****************************************************************/
	inline GLfloat	getAspectRatio() const
	{	return (GLfloat)nWidth / nHeight;	}

	inline uint_t	getTexelSize() const
	{	return nTexelSize;	}

	inline Colour	getTexel(const uint_t index) const
	{	return (this->*_getTexel)(index);	}
	inline Colour	getTexel(const uint_t nS, const uint_t nT) const
	{	return getTexel(nS + nT * nWidth);	}
	inline Colour	getTexelRepeated(const uint_t nS, const uint_t nT) const
	{	return getTexel((nS % nWidth) + (nT % nHeight) * nWidth);	}


	/****************************************************************/
	// update methods
	/****************************************************************/
	inline void		enableClamp(const bool &bEnable)
	{	bClamp = bEnable;
		if(pImageData)
			_glApplyParameters();	}
	inline void		enableMipmap(const bool &bEnable)
	{	bMipmap = bEnable;
		if(pImageData)
			_glApplyParameters();	}


	/****************************************************************/
	// OpenGL methods
	/****************************************************************/
	virtual void		glBind()		// tells OpenGL to use the texture
	{	glBindTexture(GL_TEXTURE_2D, nID);	}


	/****************************************************************/
	// operators
	/****************************************************************/
	inline Colour		operator[](const uint_t &nIndex)
	{	return getTexel(nIndex);	}
};


GRAPHICS_END
#endif
