#ifndef __GRAPHICS___GROUP_H__INCLUDED__
#define __GRAPHICS___GROUP_H__INCLUDED__

#include <list>
#include <set>

#include "SDLMutex.h"


GRAPHICS_BEGIN


template <class Item, class Base> class _Group :	public Base
{
protected:
	typedef	set<Item *>		ObjectSet;
	typedef	list<Item *>	ObjectList;


	bool		bApplyLocalTransform;

	SDLMutex	mChildren,
				mInserts,
				mRemoves,
				mRemoved;
	ObjectSet	children,
				inserted,
				removed;
	ObjectList	inserts,
				removes;


	virtual void	_load()
	{
		mChildren.P();
		for(ObjectSet::iterator it = children.begin(); it != children.end(); it++)
			(*it)->load();
		mChildren.V();
	}
	virtual void	_remove()
	{
		mChildren.P();
		for(ObjectSet::iterator it = children.begin(); it != children.end(); it++)
			(*it)->remove();
		mChildren.V();
	}

	virtual void	_glLoad()
	{
		mChildren.P();
		for(ObjectSet::iterator it = children.begin(); it != children.end(); it++)
			(*it)->glLoad();
		mChildren.V();
	}
	virtual void	_glRemove()
	{
		mChildren.P();
		for(ObjectSet::iterator it = children.begin(); it != children.end(); it++)
			(*it)->glRemove();
		mChildren.V();
	}


public:
	inline _Group(	const bool	_bApplyLocalTransform = false	) :
		bApplyLocalTransform(_bApplyLocalTransform)
	{	/* nothing to do */	}
	virtual ~_Group()
	{	_remove();	}


	// add and remove children
	inline void		add(Item *pObject)
	{	mInserts.P();
		inserts.push_back(pObject);
		mInserts.V();				}
	inline void		remove(Item *pObject)
	{	mRemoves.P();
		removes.push_back(pObject);
		mRemoves.V();				}


	virtual void	updateFinal()
	{
		tmFinal = ptmParent ? (*ptmParent) * tmLocal : tmLocal;

		if(bApplyLocalTransform)
		{
			mChildren.P();
			for(ObjectSet::iterator it = children.begin(); it != children.end(); it++)
				(*it)->updateFinal();
			mChildren.V();
		}
	}

	// advance simulation
	virtual bool	advance(float time)
	{
		bool	bContinue = false;

		mChildren.P();

		// insert new children
		mInserts.P();
		if(bApplyLocalTransform)
		{
			for(ObjectList::iterator it = inserts.begin(); it != inserts.end(); it++)
				if((children.find(*it) == children.end()) && inserted.insert(*it).second)
				{
					(*it)->glAddReference();
					(*it)->parentTransformation(&tmFinal);
					(*it)->updateFinal();
				}
		}
		else
			for(ObjectList::iterator it = inserts.begin(); it != inserts.end(); it++)
				if((children.find(*it) == children.end()) && inserted.insert(*it).second)
					(*it)->glAddReference();
		inserts.clear();
		mInserts.V();

		// removed outdated children
		mRemoves.P();
		mRemoved.P();
		for(ObjectList::iterator it = removes.begin(); it != removes.end(); it++)
			if(children.erase(*it) > 0 || inserted.erase(*it) > 0)
				if(removed.insert(*it).second)
					(*it)->glRemoveReference();
		removes.clear();
		if(removed.size())
			bContinue = true;
		mRemoved.V();
		mRemoves.V();

		// advance the simulation for all children
		Item *pObject;
		for(ObjectSet::iterator it = children.begin(); it != children.end();)
		{
			pObject = *it;

			if(!pObject->advance(time))
			{
				it = children.erase(it);

				mRemoved.P();
				if(removed.insert(pObject).second)
					pObject->glRemoveReference();
				mRemoved.V();
			}
			else
				it++;
		}
		if(children.size())
			bContinue = true;

		mChildren.V();

		// advance the simulation for all children not yet in the simulation
		for(ObjectSet::iterator it = inserted.begin(); it != inserted.end();)
		{
			pObject = *it;

			if(!pObject->advance(time))
			{
				it = inserted.erase(it);

				mRemoved.P();
				if(removed.insert(pObject).second)
					pObject->glRemoveReference();
				mRemoved.V();
			}
			else
				it++;
		}
		if(inserted.size())
			bContinue = true;

		return bContinue;
	}


	// tells all children they have to be reloaded
	virtual void	glReload()
	{
		Base::glReload();

		mChildren.P();
		for(ObjectSet::iterator it = children.begin(); it != children.end(); it++)
			(*it)->glReload();
		mChildren.V();
	}
	// update GL transformation matrix
	virtual void	glUpdate()
	{
		Base::glUpdate();

		// only one thread active - no need to lock
		if(inserted.size())
		{
			children.insert(inserted.begin(), inserted.end());
			inserted.clear();
		}

		for(ObjectSet::iterator it = removed.begin(); it != removed.end(); it++)
			if(!(*it)->getGlCount())
			{
				(*it)->glRemove();
				if((*it)->getDeleteOnRemove())
					delete *it;
			}
		removed.clear();

		for(ObjectSet::iterator it = children.begin(); it != children.end(); it++)
			(*it)->glUpdate();
	}
};


GRAPHICS_END
#endif
