#include <aquawars/Environment.h>
#include <graphics/Scene.h>


Environment::Location::Location(const Vector &position, const float &_fMaxRadius, const string &sample, const float &_fMaxVolume) :
	vPosition(position),
	fMaxRadius(_fMaxRadius),
	sndBackground(sample, true),
	fMaxVolume(_fMaxVolume),
	fVolume(0.0f),
	bStarted(false),
	bPlaying(false)
{	/* nothing to do */	}


void Environment::Location::advance(float time)
{
	if(bPlaying)
	{
		if(!bStarted)
		{
			sndBackground.start();
			bStarted = true;
		}

		fVolume += time / 5.0f;
		if(fVolume > 1.0f)
			fVolume = 1.0f;
	}
	else if(bStarted)
	{
		fVolume -= time / 5.0f;
		if(fVolume <= 0.0f)
		{
			fVolume = 0.0f;
			sndBackground.stop();
			bStarted = false;
		}
	}

	sndBackground.set_volume((int)(fVolume * fMaxVolume));
}

Environment::Environment() :
		pScene(NULL),
		nCurrent(-1),
		bSilent(false),
		fFogNear(0.001f),
		fFogFar(1.0f),
		fFogDensity(1.0f),
		fFogR(0.0f),
		fFogG(0.0f),
		fFogB(0.0f),
		fAmbR(0.0f),
		fAmbG(0.0f),
		fAmbB(0.0f)
{	/* nothing to do */	}

Environment::~Environment()
{
	for(uint_t i = 0; i < vLocations.size(); i++)
		delete vLocations[i];
}


void Environment::addEnvironment(const Environment::Vector &position, const float &fMaxRadius, const string &sample, const float &fMaxVolume)
{	vLocations.push_back(new Location(position, fMaxRadius, sample, fMaxVolume));	}


void Environment::addEnvironment(const Environment::Vector &position, const float &fMaxRadius, const string &sample, const float &fMaxVolume, const float &FogR, const float &FogG, const float &FogB,const float &FogDensity, const float &FogNear, const float &FogFar)
{	
	addEnvironment(position, fMaxRadius, sample, fMaxVolume);
	
	Location* l = vLocations[vLocations.size() - 1];
	
	l->fFogNear = FogNear;
	l->fFogFar = FogFar;
	l->fFogDensity = FogDensity;
	l->fFogR = FogR;
	l->fFogG = FogG;
	l->fFogB = FogB;
}

void Environment::addEnvironment(const Environment::Vector &position, const float &fMaxRadius, const string &sample, const float &fMaxVolume, const float &FogR, const float &FogG, const float &FogB, const float &FogDensity, const float &FogNear, const float &FogFar, const float &AmbR, const float &AmbG, const float &AmbB)
{
	addEnvironment(position, fMaxRadius, sample, fMaxVolume, FogR, FogG, FogB, FogDensity, FogNear, FogFar);
	
	Location* l = vLocations[vLocations.size() - 1];
	
	l->fAmbR = AmbR;
	l->fAmbG = AmbG;
	l->fAmbB = AmbB;	
}


void Environment::silent(const bool _silent)
{
	bSilent = _silent;
	for(uint_t i = 0; i < vLocations.size(); i++)
		vLocations[i]->stop();
}


void Environment::advance(const Vector &position, float time)
{
	if(!vLocations.size())
		return;

	float	fMinRadius,
			fRadius;
	int_t	nNew = -1;

	for(uint_t i = 0; i < vLocations.size(); i++)
	{
		const Vector	diff(position - vLocations[i]->vPosition);

		vLocations[i]->advance(time);

		fRadius = sqrt(diff.getX() * diff.getX() + diff.getZ() * diff.getZ());
		if(!vLocations[i]->fMaxRadius || fRadius <= vLocations[i]->fMaxRadius)
		{
			if(nNew < 0)
			{
                nNew = i;
				fMinRadius = fRadius;
			}
			else if(fRadius < fMinRadius)
			{
				nNew = i;
				fMinRadius = fRadius;
			}
		}
	}

	if(!bSilent && (nCurrent != nNew))
	{
		if(nCurrent >= 0)
			vLocations[nCurrent]->stop();
		if(nNew >= 0)
			vLocations[nNew]->start();

		nCurrent = nNew;
	}

	// fade to new settings
	if(time > 3.0f)
	{
		vLightSource = vLocations[nCurrent]->vLightSource;
		fTurbolence = vLocations[nCurrent]->fTurbolence;

		fFogNear = vLocations[nCurrent]->fFogNear;
		fFogFar = vLocations[nCurrent]->fFogFar;
		fFogDensity = vLocations[nCurrent]->fFogDensity;
		fFogR = vLocations[nCurrent]->fFogR;
		fFogG = vLocations[nCurrent]->fFogG;
		fFogB = vLocations[nCurrent]->fFogB;
		fAmbR = vLocations[nCurrent]->fAmbR;
		fAmbG = vLocations[nCurrent]->fAmbG;
		fAmbB = vLocations[nCurrent]->fAmbB;
	}
	else
	{
		time /= 10.0; 
		vLightSource = vLightSource * (1.0f - time) + vLocations[nCurrent]->vLightSource * time;
		fTurbolence = fTurbolence * (1.0f - time) + vLocations[nCurrent]->fTurbolence * time;

		if(vLocations[nCurrent]->fFogDensity)
		{
			fFogNear = fFogNear * (1.0f - time) + vLocations[nCurrent]->fFogNear * time;
			fFogFar = fFogFar * (1.0f - time) + vLocations[nCurrent]->fFogFar * time;
			fFogR = fFogR * (1.0f - time) + vLocations[nCurrent]->fFogR * time;
			fFogG = fFogG * (1.0f - time) + vLocations[nCurrent]->fFogG * time;
			fFogB = fFogB * (1.0f - time) + vLocations[nCurrent]->fFogB * time;
			fAmbR = fAmbR * (1.0f - time) + vLocations[nCurrent]->fAmbR * time;
			fAmbG = fAmbG * (1.0f - time) + vLocations[nCurrent]->fAmbG * time;
			fAmbB = fAmbB * (1.0f - time) + vLocations[nCurrent]->fAmbB * time;
		}
		fFogDensity = fFogDensity * (1.0f - time) + vLocations[nCurrent]->fFogDensity * time;
		if(fFogDensity < 0.00001)
			fFogDensity = 0.0f;
	}

	pScene->setFog(fFogR, fFogG, fFogB, fFogDensity, fFogNear, fFogFar, GL_LINEAR);
	pScene->setAmbientLight(fAmbR, fAmbG, fAmbB, 1.0f);
}
