#include <framework/_string.h>

#include <graphics/Font.h>
#include <graphics/_Freetype.h>


GRAPHICS_BEGIN


_Freetype						g_freetype;
ObjectManager<Font, Font::Id>	Font::manager;


void Font::_load()
{
	FT_Face			face;
	FT_BitmapGlyph	glyphs[256];

	// load the font
	face = g_freetype.loadFont(strFileName);

	try
	{	// set the size
		g_freetype.setFontSize(face, nHeight);	}
	catch(exception &)
	{	g_freetype.freeFont(face);
		throw;						}

	chars = new Character[256];
	try
	{	// load the font, character by character
		g_freetype.loadGlyphs(face, glyphs, chars);		}
	catch(exception &)
	{
		g_freetype.freeFont(face);
		g_freetype.freeGlyphs(glyphs);
		delete[] chars;

		throw;
	}

	// the font is no longer needed
	g_freetype.freeFont(face);

	// create a texture containing the glyphs
	uint_t		nHeight = 0,
				nWidth = 0,
				i;
	Character	*pChar;

	for(i = 0; i < 256; i++)
	{
		pChar = chars + i;

		nWidth += pChar->nWidth;
		if(pChar->nHeight > nHeight)
			nHeight = pChar->nHeight;
	}

	// round up to nearest power of two
	for(nTextureHeight = 1; nTextureHeight < nHeight; nTextureHeight <<=1);
	for(nTextureWidth = 1; nTextureWidth < nWidth; nTextureWidth <<=1);

	// allocate texture memory
	pTexture = new GLubyte[nTextureHeight * nTextureWidth << 1];
	if(!pTexture)
	{
		g_freetype.freeGlyphs(glyphs);
		throw OutOfMemoryError();
	}

	// copy characters into texture
	FT_Bitmap		*pBitmap;
	uint_t			nSrcX,
					nX,
					nY;

	for(i = 0, nWidth = 0; i < 256; i++)
	{
		pBitmap = &(glyphs[i]->bitmap);
		pChar = chars + i;

		// copy character data
		for(nY = 0; nY < pChar->nHeight; nY++)
			for(nSrcX = 0, nX = nWidth; nSrcX < pChar->nWidth; nSrcX++, nX++)
			{
				pTexture[(nY * nTextureWidth + nX) << 1] = pBitmap->buffer[nY * pChar->nWidth + nSrcX] ? 0xff : 0;
				pTexture[((nY * nTextureWidth + nX) << 1) + 1] = pBitmap->buffer[nY * pChar->nWidth + nSrcX];
			}

		// fill remaining space below the character
		//for(; nY < nTextureHeight; nY++)
		//	for(nX = nWidth; nX < (nWidth + pChar->nWidth); nX++)
		//		((uint16_t *)pTexture)[nY * nTextureWidth + nX] = 0;

		pChar->nTextureLeft = nWidth;
		nWidth += pChar->nWidth;
	}
	// fill remaining space right of the font
	//for(nY = 0; nY < nTextureHeight; nY++)
	//	for(nX = nWidth; nX < nTextureWidth; nX++)
	//		((uint16_t *)pTexture)[nY * nTextureWidth + nX] = 0;

	g_freetype.freeGlyphs(glyphs);
}

void Font::_remove()
{
	delete[] chars;
	delete[] pTexture;
}


void Font::_glLoad()
{
	// load texture
	glGenTextures(1, &nTextureID);

	glBindTexture(GL_TEXTURE_2D, nTextureID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(	GL_TEXTURE_2D,
					0,
					GL_LUMINANCE_ALPHA,
					nTextureWidth,
					nTextureHeight,
					0,
					GL_LUMINANCE_ALPHA,
					GL_UNSIGNED_BYTE,
					pTexture			);

	// create display lists
	nListIDBase = glGenLists(256);

	uint_t			i;
	const Character	*pChar;

	for(i = 0; i < 256; i++)
	{
		pChar = chars + i;

		// create list for a character
		glNewList(nListIDBase + i, GL_COMPILE);
			glPushMatrix();
			glTranslatef((GLfloat)pChar->nLeft, (GLfloat)(pChar->nTop) - (GLfloat)(pChar->nHeight), 0.0f);

			double	dLeft = (double)pChar->nTextureLeft / nTextureWidth,
					dRight = (double)(pChar->nTextureLeft + pChar->nWidth) / nTextureWidth,
					dTop = 0.0,
					dBottom = (double)pChar->nHeight / nTextureHeight;

			glBegin(GL_QUADS);
				glTexCoord2d(dLeft, dTop);		glVertex2f(0.0f, (GLfloat)pChar->nHeight);
				glTexCoord2d(dLeft, dBottom);	glVertex2f(0.0f, 0.0f);
				glTexCoord2d(dRight, dBottom);	glVertex2f((GLfloat)pChar->nWidth, 0.0f);
				glTexCoord2d(dRight, dTop);		glVertex2f((GLfloat)pChar->nWidth, (GLfloat)pChar->nHeight);
			glEnd();

			glPopMatrix();
			glTranslatef((GLfloat)pChar->nAdvanceX, 0.0f, 0.0f);
		glEndList();
	}
}

void Font::_glRemove()
{
	// delete display lists
	glDeleteLists(nListIDBase, 256);
	nListIDBase = 0;

	// remove texture
	glDeleteTextures(1, &nTextureID);
	nTextureID = 0;
}


Font::Font(const Id &id) :	strFileName(id.getName()),
							nHeight(id.getSize())
{
	FT_Face			face;

	// try to load the font
	face = g_freetype.loadFont(strFileName);

	try
	{	// try to set the size
		g_freetype.setFontSize(face, nHeight);	}
	catch(exception &)
	{	g_freetype.freeFont(face);
		throw;						}

	// everything looks fine - it should be possible to load the font using _load()
}


void Font::getDimensions(const std::string &str, uint_t &nWidth, GLfloat &fHeight)
{
	uint_t							nLineWidth;
	list<string>					slLines = split(str, '\n');
	list<string>::const_iterator	sliLine;

	nWidth = 0;
	for(sliLine = slLines.begin(); sliLine != slLines.end(); sliLine++)
	{
		nLineWidth = getLineWidth(*sliLine);
		if(nLineWidth > nWidth)
			nWidth = nLineWidth;
	}

	fHeight = getLineHeight() * slLines.size();
}

void Font::getDimensions(const std::string &str, GLfloat &fWidth, GLfloat &fHeight)
{
	GLfloat							fLineWidth;
	list<string>					slLines = split(str, '\n');
	list<string>::const_iterator	sliLine;

	fWidth = 0.0f;
	for(sliLine = slLines.begin(); sliLine != slLines.end(); sliLine++)
	{
		fLineWidth = (GLfloat)getLineWidth(*sliLine);
		if(fLineWidth > fWidth)
			fWidth = fLineWidth;
	}

	fHeight = getLineHeight() * slLines.size();
}

uint_t Font::getWidth(const std::string &str) const
{
	uint_t							nWidth = 0,
									nLineWidth;
	list<string>					slLines = split(str, '\n');
	list<string>::const_iterator	sliLine;

	for(sliLine = slLines.begin(); sliLine != slLines.end(); sliLine++)
	{
		nLineWidth = getLineWidth(*sliLine);
		if(nLineWidth > nWidth)
			nWidth = nLineWidth;
	}

	return nWidth;
}

uint_t Font::getLineWidth(const std::string &str) const
{
	int_t	nWidth = 0;
	uint_t	i;

	for(i = 0; i < str.length(); i++)
		nWidth += chars[str[i]].nAdvanceX;

	return nWidth;
}


void Font::_glDraw(const string &str) const
{
	list<string>					slLines = split(str, '\n');
	list<string>::const_iterator	sliLine;
	uint_t							i;
	const GLfloat					fLineHeight = nHeight * 1.3f;

	for(i = 0, sliLine = slLines.begin(); sliLine != slLines.end(); i++, sliLine++)
	{
		glPushMatrix();
		glTranslatef(0.0f, fLineHeight * i, 0.0f);
        _glDrawLine(*sliLine);
		glPopMatrix();
	}
}

void Font::_glDraw(const string &str, const GLfloat fX, const GLfloat fY, GLenum eAlignment) const
{
	list<string>					slLines = split(str, '\n');
	list<string>::const_iterator	sliLine;
	uint_t							i;
	const GLfloat					fLineHeight = nHeight * 1.3f;
	GLfloat							fYTrans;

	// calculate position of first line
	switch(eAlignment & alignYMask)
	{
	case alignYNone:
		fYTrans = 0.0f;
		break;

	case alignYLow:
		fYTrans = fY + fLineHeight * (slLines.size() - 1);
		break;
	case alignYCenter:
		fYTrans = fY + fLineHeight * (slLines.size() - 1) * 0.5f;
		break;
	case alignYHigh:
		fYTrans = fY;
		break;

	default:
		throw runtime_error("unhandled alignment");
	};

	for(i = 0, sliLine = slLines.begin(); sliLine != slLines.end(); i++, sliLine++)
        _glDrawLine(*sliLine, fX, fYTrans - (i * fLineHeight), eAlignment);
}

void Font::_glDrawLine(const string &str, const GLfloat fX, const GLfloat fY, const GLenum eAlignment) const
{
	GLfloat	fXTrans,
			fYTrans;

	glPushMatrix();

	switch(eAlignment & alignXMask)
	{
	case alignXNone:
		fXTrans = 0.0f;
		break;

	case alignXLow:
		fXTrans = fX;
		break;
	case alignXCenter:
		fXTrans = fX - getLineWidth(str) * 0.5f;
		break;
	case alignXHigh:
		fXTrans = fX - getLineWidth(str);
		break;

	default:
		throw runtime_error("unhandled alignment");
	};

	switch(eAlignment & alignYMask)
	{
	case alignYNone:
		fYTrans = 0.0f;
		break;

	case alignYLow:
		fYTrans = fY;
		break;
	case alignYCenter:
		fYTrans = fY - nHeight * 0.5f;
		break;
	case alignYHigh:
		fYTrans = fY - nHeight;
		break;

	default:
		throw runtime_error("unhandled alignment");
	};

	// render the string
	glTranslatef(fXTrans, fYTrans, 0.0f);
	glCallLists((GLsizei)str.length(), GL_UNSIGNED_BYTE, str.c_str());

	glPopMatrix();
}


void Font::glBind()
{
	// setups texture (incl. alpha blending)
	glBindTexture(GL_TEXTURE_2D, nTextureID);

	// set list base
	glListBase(nListIDBase);
}


void Font::glDrawLine(const std::string &str)
{
	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	glBindLoad();
	_glDrawLine(str);

	glPopAttrib();
}

void Font::glDrawLine(	const std::string &str,
						const GLfloat fX,
						const GLfloat fY,
						const GLenum eAlignment	)
{
	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	glBindLoad();
	_glDrawLine(str, fX, fY, eAlignment);

	glPopAttrib();
}

void Font::glDraw(const std::string &str)
{
	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	glBindLoad();
	_glDraw(str);

	glPopAttrib();
}

void Font::glDraw(	const std::string &str,
					const GLfloat fX,
					const GLfloat fY,
					GLenum eAlignment		)
{
	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	glBindLoad();
	_glDraw(str, fX, fY, eAlignment);

	glPopAttrib();
}


GRAPHICS_END
