#include <framework/Exception.h>
#include <framework/_string.h>

#include <graphics/Scene.h>
#include <graphics/Material.h>


GRAPHICS_BEGIN


void Scene::load()
{	/* nothing to do */	}


void Scene::_simulate(float time)
{
	mInserts2.P();
	if(inserts2.size())
	{
		mObjects2.P();
		for(Object2List::iterator it = inserts2.begin(); it != inserts2.end(); it++)
			if(objects2.insert(*it).second)
				(*it)->glAddReference();
		inserts2.clear();
		mObjects2.V();
	}
	mInserts2.V();

	mRemoves2.P();
	if(removes2.size())
	{
		mObjects2.P();
		for(Object2List::iterator it = removes2.begin(); it != removes2.end(); it++)
			if(objects2.erase(*it) > 0)
				if(removed.insert(*it).second)
					(*it)->glRemoveReference();
		removes2.clear();
		mObjects2.V();
	}
	mRemoves2.V();


	if(semGlUpdate.tryWait())
	{
		// only this thread is active here!
		for(ObjectList::iterator it = inserts.begin(); it != inserts.end(); it++)
			if(objects.insert(*it).second)
				(*it)->glAddReference();
		inserts.clear();
		objects.insert(inserted.begin(), inserted.end());
		inserted.clear();

		for(ObjectList::iterator it = removes.begin(); it != removes.end(); it++)
			if(objects.erase(*it) > 0)
				if(removed.insert(*it).second)
					(*it)->glRemoveReference();
		removes.clear();

		for(ObjectSet::iterator it = objects.begin(); it != objects.end(); it++)
			(*it)->glUpdate();

		semGlRender.post();
	}
	else
	{
		mObjects.P();

		mInserts.P();
		for(ObjectList::iterator it = inserts.begin(); it != inserts.end(); it++)
			if((objects.find(*it) == objects.end()) && inserted.insert(*it).second)
				(*it)->glAddReference();
		inserts.clear();
		mInserts.V();

		mRemoves.P();
		if(removes.size())
		{
			mRemoved.P();
			for(ObjectList::iterator it = removes.begin(); it != removes.end(); it++)
				if(objects.erase(*it) > 0 || inserted.erase(*it) > 0)
					if(removed.insert(*it).second)
						(*it)->glRemoveReference();
			removes.clear();
			mRemoved.V();
		}
		mRemoves.V();

		mObjects.V();
	}

	for(Object2Set::iterator it = objects2.begin(); it != objects2.end(); it++)
		if(!(*it)->advance(time))
			remove(*it);

	for(ObjectSet::iterator it = objects.begin(); it != objects.end(); it++)
		if(!(*it)->advance(time))
			remove(*it);
	for(ObjectSet::iterator it = inserted.begin(); it != inserted.end(); it++)
		if(!(*it)->advance(time))
			remove(*it);

	simulate(time);
}

void Scene::pause(float time)
{	/* nothing to do */	}

void Scene::terminate()
{
}


void Scene::run(uint32_t _nTimeMs)
{
	switch(state)
	{
	case initialising:
		_initialise();
	case loading:
		_load();
		break;

	case running:
		_simulate(_nTimeMs / 1000.0f);
		break;
	case paused:
		pause(_nTimeMs / 1000.0f);
		break;

	case terminating:
		terminate();
		cancel();
		break;

	default:
		throw Exception(getErrorMessage("unhandled state"));
		break;
	}
}


void Scene::_glReload()
{
	// read viewport properties
	glGetFloatv(GL_VIEWPORT, fViewport);

	// setup OpenGL
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glLightfv(GL_LIGHT0, GL_AMBIENT, fLightAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, fLightDiffuse);
	glLightfv(GL_LIGHT0, GL_POSITION, fLightPosition);
	glEnable(GL_LIGHT0);

	glFogi(GL_FOG_COORD_SRC, GL_FRAGMENT_DEPTH);


	// reload objects
	for(Object2Set::const_iterator it = objects2.begin(); it != objects2.end(); it++)
		(*it)->glReload();

	mObjects.P();
	for(ObjectSet::const_iterator it = objects.begin(); it != objects.end(); it++)
		(*it)->glReload();
	mObjects.V();

	if(font.isEmpty())
	{
		try
		{	font = Font::create("res/tahoma.ttf", 10);	}
		catch(Exception &e)
		{	e.report();}
		if(!font.isEmpty())
		{
			font->glLoad();
			font->glAddReference();
		}
	}
	else
		font->glReload();

	// Load caustic textures
	Object3::loadCausticTextures();
}


void Scene::glBegin2D(bool bAbsolute)
{
	// save old OpenGL settings
	glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TRANSFORM_BIT); 

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

	glLoadIdentity();
	if(bAbsolute)
		gluOrtho2D(fViewport[0], fViewport[0] + fViewport[2], fViewport[1], fViewport[1] + fViewport[3]);
	else
		gluOrtho2D(-0.5, 0.5, -0.5, 0.5);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
}

void Scene::glEnd2D()
{
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	// restore old opengl settings
	glPopAttrib();          
}


void Scene::_setupLight()
{
	glLightfv(GL_LIGHT0, GL_AMBIENT, fLightAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, fLightDiffuse);
	glLightfv(GL_LIGHT0, GL_POSITION, fLightPosition);
}

void Scene::_setupFog()
{
	glClearColor(fFogColour[0], fFogColour[1], fFogColour[2], 1.0f);

	if(bFog)
	{
		// Set Fog Color
		glFogfv(GL_FOG_COLOR, fFogColour);

		glFogf(GL_FOG_DENSITY, fFogDensity);	// How Dense Will The Fog Be
		glFogf(GL_FOG_START, fFogStart);		// Fog Start Depth
		glFogf(GL_FOG_END, fFogEnd);			// Fog End Depth

		glFogi(GL_FOG_MODE, eFogMode);			// Fog Mode
		glHint(GL_FOG_HINT, eFogHint);			// Fog Hint Value

		glEnable(GL_FOG);						// Enables GL_FOG
	}
	else
		glDisable(GL_FOG);
}


Scene::Scene(Camera *_pCamera) :	state(initialising),
									bUpdated(true),
									bReload(true),
									bFlat(false),
									bWireframe(false),
									bCaustics(true),
									bFramerate(false),
									bVFCulling(true),
									pCamera(_pCamera),
									pGlCamera(_pCamera)
{
	fLightAmbient[0] = 0.5f;
	fLightAmbient[1] = 0.5f;
	fLightAmbient[2] = 0.5f;
	fLightAmbient[3] = 1.0f;

	fLightDiffuse[0] = 1.0f;
	fLightDiffuse[1] = 1.0f;
	fLightDiffuse[2] = 1.0f;
	fLightDiffuse[3] = 1.0f;

	fLightPosition[0] = 0.0f;
	fLightPosition[1] = 1.0f;
	fLightPosition[2] = 0.0f;
	fLightPosition[3] = 0.0f;

	resetFog();

	semGlUpdate.post();
}

Scene::~Scene()
{	/* nothing to do */	}


void Scene::clear()
{
	mInserts2.P();
	inserts2.clear();
	mInserts2.V();
	mInserts.P();
	inserts.clear();
	mInserts.V();

	mObjects2.P();
	for(Object2Set::iterator it = objects2.begin(); it != objects2.end(); it++)
		remove(*it);
	mObjects2.V();
	mObjects.P();
	for(ObjectSet::iterator it = objects.begin(); it != objects.end(); it++)
		remove(*it);
	mObjects.V();
	for(ObjectSet::iterator it = inserted.begin(); it != inserted.end(); it++)
		remove(*it);

	resetFog();
	resetAmientLight();
}


void Scene::getCollisions(list<_Object3 *> &lCollisions, _Object3 *pObject)
{
	mObjects.P();
	for(ObjectSet::iterator it = objects.begin(); it != objects.end(); it++)
		if(*it != pObject && (*it)->collides(*pObject))
			lCollisions.push_back(*it);
	mObjects.V();
	for(ObjectSet::iterator it = inserted.begin(); it != inserted.end(); it++)
		if(*it != pObject && (*it)->collides(*pObject))
			lCollisions.push_back(*it);
}

void Scene::getObjectsInRadius(list<_Object3 *> &lObjects, _Object3 *pObject, const float &radius)
{
	mObjects.P();
	for(ObjectSet::iterator it = objects.begin(); it != objects.end(); it++)
		if(*it != pObject && (Vector((*it)->getPosition() - pObject->getPosition()).magnitude() <= radius))
			lObjects.push_back(*it);
	mObjects.V();
	for(ObjectSet::iterator it = inserted.begin(); it != inserted.end(); it++)
		if(*it != pObject && (Vector((*it)->getPosition() - pObject->getPosition()).magnitude() <= radius))
			lObjects.push_back(*it);
}


void Scene::setAmbientLight(const FloatColour &c)
{
	fLightAmbient[0] = c.getRed();
	fLightAmbient[1] = c.getGreen();
	fLightAmbient[2] = c.getBlue();
	fLightAmbient[3] = c.getAlpha();
}

void Scene::setAmbientLight(const GLfloat &r, const GLfloat &g, const GLfloat &b, const GLfloat &a)
{
	fLightAmbient[0] = r;
	fLightAmbient[1] = g;
	fLightAmbient[2] = b;
	fLightAmbient[3] = a;
}

void Scene::resetAmientLight()
{
	fLightAmbient[0] = 0.0f;
	fLightAmbient[1] = 0.0f;
	fLightAmbient[2] = 0.0f;
	fLightAmbient[3] = 1.0f;
}


void Scene::setLight(const FloatColour &c, const Vector &direction)
{
	fLightDiffuse[0] = c.getRed();
	fLightDiffuse[1] = c.getGreen();
	fLightDiffuse[2] = c.getBlue();
	fLightDiffuse[3] = c.getAlpha();

	fLightPosition[0] = -direction.getX();
	fLightPosition[1] = -direction.getY();
	fLightPosition[2] = -direction.getZ();
	fLightPosition[3] = 0.0f;
}

void Scene::setLight(	const GLfloat &r, const GLfloat &g, const GLfloat &b, const GLfloat &a,
						const GLfloat &x, const GLfloat &y, const GLfloat &z					)
{
	fLightDiffuse[0] = r;
	fLightDiffuse[1] = g;
	fLightDiffuse[2] = b;
	fLightDiffuse[3] = a;

	fLightPosition[0] = -x;
	fLightPosition[1] = -y;
	fLightPosition[2] = -z;
	fLightPosition[3] = 0.0f;
}


void Scene::setFog(	const GLfloat &red, const GLfloat &green, const GLfloat &blue,
					const GLfloat &density, const GLfloat &start, const GLfloat &end,
					const GLenum &mode, const GLenum &hint								)
{
	fFogColour[0] = red;
	fFogColour[1] = green;
	fFogColour[2] = blue;
	fFogColour[3] = 1.0f;

	fFogDensity = density;
	fFogStart = start;
	fFogEnd = end;

	eFogMode = mode;
	eFogHint = hint;

	bFog = true;
}


void Scene::glRender()
{
	// set up camera
	pGlCamera = pCamera;
	if(!pGlCamera)
		return;

	// reconfigure OpenGL
	if(bReload)
	{	_glReload();
		bReload = false;	}

	if(bFlat)
		glShadeModel(GL_FLAT);
	else
		glShadeModel(GL_SMOOTH);

	if(bWireframe)
		glPolygonMode(GL_FRONT, GL_LINE);
	else
		glPolygonMode(GL_FRONT, GL_FILL);

	_setupFog();
	_setupLight();


	// begin rendering
	semGlRender.wait();


	Vector		vCamera(pCamera->getPositionF());

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	pGlCamera->glLook();

	// create a list of visible objects
	mObjects.P();
	if(bVFCulling)
		for(ObjectSet::iterator it = objects.begin(); it != objects.end(); it++)
			(*it)->glGetVisible(rm, pGlCamera, vCamera);
	else
		for(ObjectSet::iterator it = objects.begin(); it != objects.end(); it++)
			(*it)->glGetAll(rm, vCamera);
	mObjects.V();

	//enalbe lighting and texturing
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);

	// render visible 3D objects
	const uint_t	nVisible = (uint_t)rm.size();
	for(RenderMap::const_iterator it = rm.begin(); it != rm.end(); it++)
		it->second->glRender();
	rm.clear();


	// render all 2D objects
	glBegin2D(true);

	mObjects2.P();
	for(Object2Set::iterator it = objects2.begin(); it != objects2.end(); it++)
	{
		(*it)->glUpdate();
		(*it)->glRender();
	}
	mObjects2.V();

	framerate.addFrame(SDL_GetTicks());
	if(bFramerate)
	{
		char	str[256];

		sprintf(str, "%u objects visible\n%2.1f fps", nVisible, framerate.getFrameRate());

		glColor4f(1.0f, 1.0f, 1.0f, 0.75f);
		font->glDraw("AquaWars - Return of Luzi", 0.0f, getHeight(), Font::alignLowHigh);
		font->glDraw(str, getWidth(), getHeight(), Font::alignHighHigh);
	}

	glEnd2D();


	// remove unused objects
	mRemoved.P();
	for(RemovedSet::iterator it = removed.begin(); it != removed.end(); it++)
		if(!(*it)->getGlCount())
		{
			(*it)->glRemove();
			if((*it)->getDeleteOnRemove())
				delete *it;
		}
	removed.clear();
	mRemoved.V();


	semGlUpdate.post();
}


GRAPHICS_END
