/****************************************************************/
// include files
/****************************************************************/
#include <graphics/Texture.h>
#include <graphics/TGATexture.h>


GRAPHICS_BEGIN


Texture::Manager	Texture::manager;
bool				Texture::bGlobalMipmap = true;


Texture::Id::Id(	const string &_file,
					const bool _crop,
					const bool _clamp,
					const bool _mipmap	) :
	classId(_getExtension(_file)),
	file(_file),
	crop(_crop),
	clamp(_clamp),
	mipmap(_mipmap)
{	/* nothing to do */	}


void Texture::Manager::_init()
{
	pTGAAllocator =  new TGATexture::Allocator;
	if(!pTGAAllocator)
		throw OutOfMemoryError();

	allocator.add(Id::ClassId("tga"), pTGAAllocator);
}

Texture::Manager::~Manager()
{	delete pTGAAllocator;	}


/****************************************************************/
// Texture implementation
/****************************************************************/
void Texture::_init_type()
{
	switch(nType)
	{
	case GL_RED:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_RED;
		nTexelSize = 1;
		break;
	case GL_GREEN:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_GREEN;
		nTexelSize = 1;
		break;
	case GL_BLUE:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_BLUE;
		nTexelSize = 1;
		break;
	case GL_ALPHA:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_ALPHA;
		nTexelSize = 1;
		break;
	case GL_LUMINANCE:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_LUMINANCE;
		nTexelSize = 1;
		break;
	case GL_LUMINANCE_ALPHA:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_LUMINANCE_ALPHA;
		nTexelSize = 2;
		break;
	case GL_RGB:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_RGB;
		nTexelSize = 3;
		break;
	case GL_BGR_EXT:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_BGR_EXT;
		nTexelSize = 3;
		break;
	case GL_RGBA:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_RGBA;
		nTexelSize = 4;
		break;
	case GL_BGRA_EXT:
		_getTexel = (fnGetTexel)&Texture::_getTexel_GL_BGRA_EXT;
		nTexelSize = 4;
		break;

	default:
		_getTexel = (fnGetTexel)&Texture::_getTexel_unknown;
		nTexelSize = 0;
		break;
	};
}


void Texture::_remove()
{
	delete[] pImageData;
	pImageData = NULL;
}


void Texture::_glRemove()
{	glDeleteTextures(1, &nID);	}

void Texture::_glApplyParameters() const
{
	if(bClamp)
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	}

	if(bGlobalMipmap && bMipmap)
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		gluBuild2DMipmaps(GL_TEXTURE_2D, 3, nWidth, nHeight, nType, GL_FLOAT, pImageData);
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D(	GL_TEXTURE_2D,
						0,
						nType,
						nWidth,
						nHeight,
						0,
						nType,
						GL_FLOAT,
						pImageData		);
	}
}


Texture::Colour Texture::_getTexel_GL_RED(const uint_t index) const
{	return Texture::Colour(pImageData[index], 0, 0);		}
Texture::Colour Texture::_getTexel_GL_GREEN(const uint_t index) const
{	return Texture::Colour(0, pImageData[index], 0);		}
Texture::Colour Texture::_getTexel_GL_BLUE(const uint_t index) const
{	return Texture::Colour(0, 0, pImageData[index], 0);	}
Texture::Colour Texture::_getTexel_GL_ALPHA(const uint_t index) const
{	return Texture::Colour(0, 0, 0, pImageData[index]);	}
Texture::Colour Texture::_getTexel_GL_LUMINANCE(const uint_t index) const
{	return Texture::Colour(pImageData[index], pImageData[index], pImageData[index]);	}
Texture::Colour Texture::_getTexel_GL_LUMINANCE_ALPHA(const uint_t index) const
{
	return Texture::Colour(	pImageData[index << 1],
							pImageData[index << 1],
							pImageData[index << 1],
							pImageData[(index << 1) + 1]	);
}
Texture::Colour Texture::_getTexel_GL_RGB(const uint_t index) const
{
	return Texture::Colour(	pImageData[index * 3],
							pImageData[index * 3 + 1],
							pImageData[index * 3 + 2]	);
}
Texture::Colour Texture::_getTexel_GL_BGR_EXT(const uint_t index) const
{
	return Texture::Colour(	pImageData[index * 3 + 2],
							pImageData[index * 3 + 1],
							pImageData[index * 3]		);
}
Texture::Colour Texture::_getTexel_GL_RGBA(const uint_t index) const
{
	return Texture::Colour(	pImageData[index << 2],
							pImageData[(index << 2) + 1],
							pImageData[(index << 2) + 2],
							pImageData[(index << 2) + 3]	);
}
Texture::Colour Texture::_getTexel_GL_BGRA_EXT(const uint_t index) const
{
	return Texture::Colour(	pImageData[(index << 2) + 2],
							pImageData[(index << 2) + 1],
							pImageData[index << 2],
							pImageData[(index << 2) + 3]	);
}
Texture::Colour Texture::_getTexel_unknown(const uint_t index) const
{	throw NotImplementedError();	}


Texture::Texture(	const bool bClamp,
					const bool bMipmap		) :	GlItem(),
												bClamp(bClamp),
												bMipmap(bMipmap)
{
	nID = 0;
	pImageData = NULL;
}

Texture::~Texture()
{
	remove();
	glRemove();
}


uint_t Texture::powerOf2(const uint_t upperBound)
{
	uint_t	result;
	for(result = 1 << ((sizeof(uint_t) << 3) - 1); result > upperBound; result >>= 1);
	return result;
}


void Texture::_glLoad()
{
	glGenTextures(1, &nID);
	glBind();
	_glApplyParameters();
}


GRAPHICS_END
