#include <aquawars/Vehicle.h>

#include <graphics/Map.h>
#include <graphics/ParticleList.h>
#include <graphics/Scene.h>

#include <aquawars/ObjectTypes.h>


Sound	Vehicle::sndExplosion0("res/sounds/explosion0.wav"),
		Vehicle::sndExplosion1("res/sounds/explosion1.wav"),
		Vehicle::sndExplosion2("res/sounds/explosion2.wav"),
		Vehicle::sndExplosion3("res/sounds/explosion3.wav"),
		Vehicle::sndExplosion4("res/sounds/explosion4.wav"),
		Vehicle::sndExplosion5("res/sounds/explosion5.wav"),
		Vehicle::sndExplosion6("res/sounds/explosion6.wav"),
		Vehicle::sndExplosion7("res/sounds/explosion7.wav");


void Vehicle::_load()
{
	Group3::_load();
	pModel->load();
}

void Vehicle::_remove()
{
	Group3::_remove();
	//pModel->remove();
}


Vehicle::Vehicle(const string &strModel, const float &_fMaxEnginePower, const float &_fMaxArmour, const float &_fScale) :
	Group3(true, true),
	fMaxEnginePower(_fMaxEnginePower),
	fMaxArmour(_fMaxArmour),
	fEnginePower(0.0f),
	nWeaponIndex(0),
	bFire(false),
	bDestroyed(false),
	bGod(false),
	fSpeed(0.0f),
	fPhi(0.0f),
	fPhiDiff(0.0f),
	fPhiDiffNew(0.0f),
	fTheta(0.0f),
	fThetaDiff(0.0f),
	fThetaDiffNew(0.0f),
	fCloudCoolDown(0.0f),
	fBubbleCoolDown(0.0f)
{
	type(c_vehicle, vehicle);
	boundingSphere(_fScale);

	fArmour[0] = fMaxArmour;
	fArmour[1] = fMaxArmour;
	fArmour[2] = fMaxArmour;
	fArmour[3] = fMaxArmour;
	fArmour[4] = fMaxArmour;
	fArmour[5] = fMaxArmour;

	pModel = new Model(strModel);

	pModel->deleteOnRemove(true);
	pModel->reset();
	pModel->postScale(_fScale, _fScale, _fScale);
	pModel->postRotateY((float)PI);
	pModel->postRotateX((float)PI / 2.0f);
	pModel->updateFinal();

	add(pModel);
}


void Vehicle::refillWeapons()
{
	for(size_t i = 0; i < weapons.size(); i++)
		weapons[i].refill();
}


void Vehicle::repair()
{
	for(uint_t i = 0; i < 6; i++)
		fArmour[i] = fMaxArmour;
}


void Vehicle::onHit(const Vector &where, const float &damage)
{
	if(bGod)
		return;

	float	x = where.getX(),
			y = where.getY(),
			z = where.getZ();

	if(boundaryType == box)
	{
		x /= pBoundary[0];
		y /= pBoundary[1];
		z /= pBoundary[2];
	}

	ArmourSegment	as;

	if(fabs(x) > fabs(y))
	{
		if(fabs(x) > fabs(z))
			// x
			as = x > 0.0f ? left : right;
		else
			// z
			as = z < 0.0f ? forward : backward;
	}
	else
	{
		if(fabs(y) > fabs(z))
			// y
			as = y < 0.0f ? down : up;
		else
			// z
			as = z < 0.0f ? forward : backward;
	}

	fArmour[as] -= damage;
	if(fArmour[as] <= 0.0f)
		bDestroyed = true;
}

void Vehicle::onTargetDestroyed(const Vehicle *pTarget)
{	/* nothing to do */	}


bool Vehicle::advance(float fTime)
{
	Group3::advance(fTime);

	fCloudCoolDown -= fTime;
	fBubbleCoolDown -= fTime;

	if(bDestroyed)
	{
		if(pScene)
		{
			spawn(ParticleList::createExplosion3D(10.0f, 10.0f, 4.5f, 250, 150, getPosition()));

			switch(rand() % 8)
			{
			case 0:
				sndExplosion0.start();
				break;
			case 1:
				sndExplosion1.start();
				break;
			case 2:
				sndExplosion2.start();
				break;
			case 3:
				sndExplosion3.start();
				break;
			case 4:
				sndExplosion4.start();
				break;
			case 5:
				sndExplosion5.start();
				break;
			case 6:
				sndExplosion6.start();
				break;
			case 7:
				sndExplosion7.start();
				break;
			}
		}

		return false;
	}

	fSpeed *= pow(0.9975f, fTime * 60.0f);
	fSpeed += fMaxEnginePower * fEnginePower * fTime;

	fPhiDiff = 0.9f * fPhiDiff + 0.1f * fPhiDiffNew;
	fPhiDiffNew = 0.0f;
	fThetaDiff = 0.9f * fThetaDiff + 0.1f * fThetaDiffNew;
	fThetaDiffNew = 0.0f;
	fPhi += fPhiDiff;
	fTheta += fThetaDiff;
	
	if(fPhi > (GLfloat)PI / 4.0f)
		fPhi = (GLfloat)PI / 4.0f;
	else if(fPhi < (GLfloat)PI / -4.0f)
		fPhi = (GLfloat)PI / -4.0f;

	if(fTheta > (GLfloat)PI)
		fTheta -= (GLfloat)PI * 2.0f;
	else if(fTheta < -(GLfloat)PI)
		fTheta += (GLfloat)PI * 2.0f;

	// position sub
	reset();
	postRotateY(fTheta);
	postRotate(getLocalForward(), -fThetaDiff * (float)PI / 1.5f);
	postRotate(getLocalStrafe(), fPhi);
	postTranslate(getPosition() + getForward() * fSpeed * fTime);
	updateFinal();

	// check for a collision with the terrain
	if(pScene)
	{
		list<_Object3 *>	objects;
		pScene->getCollisions(objects, this);
		for(list<_Object3 *>::iterator it = objects.begin(); it != objects.end(); it++)
		{
			switch((*it)->getTypeID())
			{
			case terrain:
				{
					Map		*pMap = (Map *)((*it)->getTypeData());

					if(pMap)
					{
						Vector	pos(getPosition());
						float	h = pMap->heightAt(pos.getX(), pos.getZ());

						if(h > pos.getY())
						{
							if(fCloudCoolDown <= 0.0f)
							{
								spawn(ParticleList::createCloudExplosion(6, getPosition(), 5.0f));
								fCloudCoolDown = 0.1f;
							}
							postTranslate(0.0f, h - pos.getY(), 0.0f);
							updateFinal();
							fSpeed *= 0.9f;
						}
					}
					break;
				}

			case wallx:
				{
					if(getForward().getZ() > 0.0f)
						postTranslate(0.0f, 0.0f, -0.5f);
					else
						postTranslate(0.0f, 0.0f, 0.5f);
					updateFinal();

					//fSpeed.z() = 0.0f;
					fSpeed *= 0.9f;
					break;
				}

			case wallz:
				{
					if(getForward().getZ() > 0.0f)
						postTranslate(-0.5f, 0.0f, 0.0f);
					else
						postTranslate(0.5f, 0.0f, 0.0f);
					updateFinal();

					//vSpeed.x() = 0.0f;
					fSpeed *= 0.9f;
					break;
				}

			case vehicle:
			case player:
				((Vehicle *)(*it))->onHit(getPosition(), 100.0f);
				if(((Vehicle *)(*it))->isDestroyed())
					onTargetDestroyed(this);
				onHit((*it)->getPosition(), 100.0f);
				if(isDestroyed())
					onTargetDestroyed((Vehicle *)(*it));
				break;
			}
		}
	}

	for(vector<Weapon>::iterator it = weapons.begin(); it != weapons.end(); it++)
		it->advance(fTime);
	if(bFire && weapons.size())
		spawn(weapons[nWeaponIndex].fire(getPosition(), getForward(), fSpeed));

	// Bubbles
	if(fBubbleCoolDown <= 0.0f)
	{
		spawn(ParticleList::createBubbleExplosion(1, getPosition()-getForward()*2, 1.25f));
		fBubbleCoolDown = 0.1f;
	}

	return true;
}
